/*
 * Tile management functions.
 */

#include <nds.h>
#include "hack.h"

#include "bmp.h"
#include "ppm-lite.h"
#include "font-bdf.h"

#include "nds_win.h"
#include "nds_util.h"
#include "nds_map.h"
#include "nds_gfx.h"

#define FONT_FILE_NAME          "map.bdf"
#define FONT_PALETTE_NAME       "map.pal"

#define MAP_PALETTE_BASE        32

#define TILE_WIDTH		iflags.wc_tile_width
#define TILE_HEIGHT		iflags.wc_tile_height
#define TILE_FILE		iflags.wc_tile_file

#define MAP_BASE BG_MAP_BASE(8)
#define TILE_BASE BG_TILE_BASE(6)

#define MAX_TILE_SLOTS          512

#define MINIMAP_X               4

#define NDS_LOAD_TILE(tileidx, glyph, x, y) ((TILE_FILE == NULL) ? nds_load_text_tile(tileidx, glyph, x, y) : nds_load_graphics_tile(tileidx, glyph, x, y))
#define NDS_INIT_MAP(pal, pallen) ((TILE_FILE == NULL) ? nds_init_text_map(pal, pallen) : nds_init_tiled_map(pal, pallen))

#define ROUND_UP(val) ( (((val) & 0x07) == 0) ? (val) : (((val / 8) + 1) * 8) )

#define c2(a,i)		(RGB15((a[i+2]>>3),(a[i+1]>>3),(a[i]>>3)))

typedef struct {
  int glyph;
  int tile;
  int ch;
  int colour;
  int special;
  long last_used;     /* This is in game time (moves)          */
} tile_cache_entry_t;

/* Tile-based graphics specific variables */

bmp_t tiles;

int width_in_tiles;
int height_in_tiles;

/* Text-based graphics variables */

struct font *map_font;
struct ppm *text_img;
int font_char_w;
int font_char_h;

/* General rendering variables */

u16 *tile_ram = (u16 *)BG_TILE_RAM(6);
u16 *map_ram = (u16 *)BG_MAP_RAM(8);

tile_cache_entry_t *tile_cache;
int num_cache_entries;

int tile_width;
int tile_height;

int map_width;
int map_height;

int cx, cy;
int pet_count;

/* Sprite-specific variables (for tile mode) */

u16 *oam_ram = OAM;
tOAM oam_shadow;

/*
 * Initialize our tile cache.
 */
void nds_init_tile_cache()
{
  int i;

  num_cache_entries = (MAX_TILE_SLOTS - 1) / (tile_width * tile_height);

  tile_cache = (tile_cache_entry_t *)malloc(sizeof(tile_cache_entry_t) * num_cache_entries);

  for (i = 0; i < num_cache_entries; i++) {
    tile_cache[i].glyph = -1;
  }
}

/*
 * Returns the cache slot number for the given glyph, if one exists.
 */
int nds_find_cache_slot(int glyph, int gx, int gy)
{
  int i;
  int ch, colour;
  unsigned int special;
  int tile = glyph2tile[glyph];

  mapglyph(glyph, &ch, &colour, &special, gx, gy);

  for (i = 0; i < num_cache_entries; i++) {
    if ((tile_cache[i].glyph == glyph) && (tile_cache[i].colour == colour) && 
        (tile_cache[i].special == special) && (tile_cache[i].ch == ch) &&
        (tile_cache[i].tile == tile)) {
      return i;
    }
  }

  return -1;
}

/*
 * Allocate a cache slot for the given glyph.  Note, this may evict an entry
 * if the cache is full.
 */
int nds_allocate_cache_slot(int glyph, int gx, int gy)
{
  int cache_slot = -1;
  int i;

  int oldest_diff = 0;
  int oldest;

  int ch, colour;
  unsigned int special;

  for (i = 0; i < num_cache_entries; i++) {
    int diff;

    if (tile_cache[i].glyph < 0) {
      cache_slot = i;

      break;
    } else if (tile_cache[i].last_used < 0) {
      continue;
    }

    diff = moves - tile_cache[i].last_used;

    if (diff > oldest_diff) {
      oldest = i;
      oldest_diff = diff;
    }
  }

  if (cache_slot < 0) {
    cache_slot = oldest;
  }

  mapglyph(glyph, &ch, &colour, &special, gx, gy);

  tile_cache[cache_slot].glyph = glyph;
  tile_cache[cache_slot].colour = colour;
  tile_cache[cache_slot].special = special;
  tile_cache[cache_slot].ch = ch;
  tile_cache[cache_slot].tile = glyph2tile[glyph];

  return cache_slot;
}

/*
 * Copy a tile from our tile image, loaded previously, into tile RAM.  In the
 * case of tiles with a width or height larger than 8 pixels, this will result
 * in multiple tile positions being occupied in video RAM.
 */
void nds_load_graphics_tile(int tile_idx, int glyph, int gx, int gy) 
{
  int i, j, x;
  int width, height, bpp, row_bytes;
  int bmp_tile_x, bmp_tile_y;
  u8 *bmp_row_start;
  u16 *tile_row_start;

  /* Next, get a few fields from the BMP headers that we'll need. */

  width = bmp_width(&tiles);
  height = bmp_height(&tiles);
  bpp = bmp_bpp(&tiles);

  /*
   * Compute the number of bytes in each row of 8-pixel tiles.
   *
   * This works explicitely because each tile component is 8 pixels wide.
   * Thus, the number of bytes in a row is (8 * (bpp / 8)), which just
   * ends up being bpp.
   */

  row_bytes = tile_width * bpp;

  /* Now calculate the pointer which points to the start of the BMP row */

  bmp_tile_y = glyph2tile[glyph] / width_in_tiles;
  bmp_tile_x = width_in_tiles - glyph2tile[glyph] % width_in_tiles;

  bmp_row_start = tiles.bitmap + tiles.bitmap_length - 
                  bmp_tile_y * TILE_HEIGHT * width_in_tiles * row_bytes - 
                  bmp_tile_x * row_bytes ;

  for (j = 0; j < tile_height; j++) {
    int y;

    for (y = 0; y < 8; y++) {
      for (i = 0; i < tile_width; i++) {
        tile_row_start = tile_ram + (tile_idx + j * tile_width + i) * bpp * 8 / 2 + y * bpp / 2; 

        /* Again, this works because 8 * (bpp / 8) == bpp */

        for (x = 0; x < bpp; x += 2, bmp_row_start += 2) {
          if (bpp == 4) {
            u16 a, b, c, d;

            a = (bmp_row_start[0] & 0xF0) >> 4;
            b = (bmp_row_start[0] & 0x0F);
            c = (bmp_row_start[1] & 0xF0) >> 4;
            d = (bmp_row_start[1] & 0x0F);

            tile_row_start[x / 2] = (d << 12) | (c << 8) | (b << 4) | (a << 0);
          } else {
            tile_row_start[x / 2] = (bmp_row_start[1] << 8) |
                                     bmp_row_start[0];
          }
        }
      }

      bmp_row_start -= row_bytes * width_in_tiles + row_bytes;
    }
  }
}

/*
 * Render the glyph into a tile.
 */
void nds_load_text_tile(int tile_idx, int glyph, int gx, int gy) 
{
  int ch, color;
  unsigned int special;
  char tmp[BUFSZ];
  int tile_x, y, i;
  u16 *tile_ptr;
  unsigned char *img_data;
  int black = MAP_COLOUR(CLR_BLACK);

  /* Alright, now convert the glyph to a character */

  mapglyph(glyph, &ch, &color, &special, gx, gy);

  /* Mmm... hacky... */

  if (color == CLR_BLACK) {
    color = CLR_BLUE;
  }

  if (((special & MG_PET) && iflags.hilite_pet) ||
      ((special & MG_DETECT) && iflags.use_inverse)) {
    sprintf(tmp, "\e[7m%c", ch);
  } else {
    sprintf(tmp, "%c", ch);
  }

  /* Now draw the character to a PPM image... yes, this is inefficient :) */

  clear_ppm(text_img, black);

  draw_string(map_font, tmp, text_img, 
              text_img->width / 2 - font_char_w / 2, 
              text_img->height / 2 - font_char_h / 2, 
              -1, -1);

  img_data = (unsigned char *)text_img->bitmap;
  tile_ptr = tile_ram + tile_idx * 64 / 2;

  /* Now copy the contents of the PPM to tile RAM */

  for (y = 0; y < text_img->height; y++) {
    for (tile_x = 0; tile_x < tile_width; tile_x++) {
      int tile_y = y & 0xF0;
      int tile_row = y & 0x0F;
      u16 *row_ptr = tile_ptr + 
                     ((tile_y * tile_width + tile_x) * 64 +
                      tile_row * 8) / 2;

      for (i = 0; i < 4; i++, img_data += 2) {
        u8 c0 = ((img_data[0] != black) ? color : 0) + 1;
        u8 c1 = ((img_data[1] != black) ? color : 0) + 1;

        row_ptr[i] = (c1 << 8) |
                     c0;
      }
    }
  }
}

int nds_load_tile(int glyph, int gx, int gy)
{
  int cache_slot;

  if ((cache_slot = nds_find_cache_slot(glyph, gx, gy)) < 0) {
    cache_slot = nds_allocate_cache_slot(glyph, gx, gy);

    NDS_LOAD_TILE((cache_slot + 1) * (tile_width * tile_height), glyph, gx, gy);
  }

  tile_cache[cache_slot].last_used = moves;

  return (cache_slot + 1) * (tile_width * tile_height);
}

/*
 * Plot the specified tile to the map.  Note, this represents the tile index
 * as present in the BMP file, *not* the tile RAM index.
 */
void nds_draw_tile(nds_map_t *map, int glyph, int x, int y, int gx, int gy)
{
  int midx, tidx;
  int i, j;
  int palette;

  if (glyph < 0) {
    tidx = 0x0000;
  } else {
    tidx = nds_load_tile(glyph, gx, gy);
  }

  /*
   * midx is the starting map index.
   * tidx is the starting tile index.
   *
   * We do a bit of tricky looping here to keep things fast.  midx points to
   * the starting map index.  Each time we complete a row iteration, we move
   * midx ahead by map_width, so it points to the start of the next row.  We
   * then use that to loop through the row copying the tile indices over.
   *
   * Meanwhile, tidx starts off as the initial tile index in RAM.  Because
   * the tiles are ordered left-to-right, top-to-bottom, we can just
   * increment the counter as we populate the map.
   */

  if (TILE_FILE == NULL) {
    switch (iflags.cursor) {
      case 0:
        palette = ((gx == map->cx) && (gy == map->cy)) ? 3 : 2;
        break;

      case 1:
        palette = (((gx != u.ux) || (gy != u.uy)) && (gx == map->cx) && (gy == map->cy)) ? 3 : 2;
        break;

      case 2:
        palette = 2;
        break;

      default:
        break;
    }
  } else {
    palette = 2;
  }

  midx = (y * tile_height) * 32 + x * tile_width;

  for (j = 0; j < tile_height; j++, midx += 32) {
    for (i = 0; i < tile_width; i++) {
      map_ram[midx + i] = tidx | (palette << 12); 

      if (tidx > 0) {
        tidx++;
      }
    }
  }

  /* 
   * Yeah, this is a weird place for this, BUT... here we put in a sprite
   * cursor if this is a pet and we're in tile mode.
   */
  if ((TILE_FILE != NULL) && iflags.hilite_pet) {
    int ch, color;
    unsigned int special;

    mapglyph(glyph, &ch, &color, &special, gx, gy);

    if (special & MG_PET) {
      oam_shadow.spriteBuffer[++pet_count].posX = x * TILE_WIDTH;
      oam_shadow.spriteBuffer[pet_count].posY = y * TILE_HEIGHT;
      oam_shadow.spriteBuffer[pet_count].isHidden = 0;
    }
  }
}

/*
 * Initialize the graphical tile map.
 */
int nds_init_tiled_map(u16 *palette, int *pallen)
{
  char *fname = TILE_FILE;
  int i;

  tile_width = TILE_WIDTH / 8;
  tile_height = TILE_HEIGHT / 8;

  /* Now load the tiles into memory */

  if (bmp_read(fname, &tiles) < 0) {
    return -1;
  }

  /* Compute the width and height of our image, in tiles */

  width_in_tiles = bmp_width(&tiles) / TILE_WIDTH;
  height_in_tiles = bmp_height(&tiles) / TILE_HEIGHT;

  /* Alright, file loaded, let's copy over the palette */

  *pallen = tiles.palette_length;

  for (i = 0; i < tiles.palette_length; i++) {
    u16 val = RGB15((tiles.palette[i].r >> 3),
                    (tiles.palette[i].g >> 3),
                    (tiles.palette[i].b >> 3));

    palette[i] = val;
  }

  return bmp_bpp(&tiles);
}

/*
 * Initialize the textual map.
 */
int nds_init_text_map(u16 *palette, int *pallen)
{
  int img_w, img_h;
  int i;
  int palcnt;

  if ((map_font = read_bdf(FONT_FILE_NAME)) == NULL) {
    iprintf("Unable to open '%s'\n", FONT_FILE_NAME);

    return -1;
  }

  if ((palcnt = nds_load_palette(FONT_PALETTE_NAME, palette + 1)) < 0) {
    return -1;
  }

  /* Generate our inverse palette */

  for (i = 1; i < palcnt + 1; i++) {
    palette[i + 256] = palette[i] ^ 0x7FFF;
  }

  *pallen = 512;

  /* Now figure out our dimensions */

  text_dims(map_font, "#", &font_char_w, &font_char_h);

  img_w = ROUND_UP(font_char_w);
  img_h = ROUND_UP(font_char_h);

  text_img = alloc_ppm(img_w, img_h);

  tile_width = img_w / 8;
  tile_height = img_h / 8;

  /* Lastly, create out "blank" tile */

  for (i = 0; i < tile_width * tile_height * 32; i++) {
    tile_ram[i] = 0x0101;
  }

  return 8;
}

void nds_clear_sprites()
{
  int i;

  for (i = 0; i < 128; i++) {
    oam_shadow.spriteBuffer[i].isHidden = 1;
  }
}

/*
 * Get the user sprite set up and drawn.
 */
void nds_draw_graphics_cursor(int x, int y)
{
  if ((x < 0) || (y < 0) || (x > map_width) || (y > map_height) ||
      (! iflags.cursor)) {

    return;
  } 

  oam_shadow.spriteBuffer[0].posX = x * TILE_WIDTH;
  oam_shadow.spriteBuffer[0].posY = y * TILE_HEIGHT;
  oam_shadow.spriteBuffer[0].isHidden = 0;
}

void nds_render_sprite(int bpp, int spr_size, int spr_num, int colour)
{
  u16 *spr_gfx_ram = SPRITE_GFX;
  int x, y;

  spr_gfx_ram += (spr_size * spr_size * bpp * 8 / 2) * spr_num;

  for (y = 0; y < TILE_HEIGHT; y++) {
    for (x = 0; x < TILE_WIDTH; x++) {
      int tidx;
      u16 *tptr;

      int tx = x & 0x07;
      int ty = y & 0x07;

      if ((x != 0) && (x != (TILE_WIDTH - 1)) && 
          (y != 0) && (y != (TILE_HEIGHT - 1))) {
        continue;
      }

      if ( ((x == 0) || (x == (TILE_WIDTH - 1))) &&
           ((y > 2) && (y < (TILE_HEIGHT - 3))) ) {
        continue;
      }

      if ( ((y == 0) || (y == (TILE_HEIGHT - 1))) &&
           ((x > 2) && (x < (TILE_WIDTH - 3))) ) {
        continue;
      }

      tidx = (y / 8) * spr_size + x / 8;
      tptr = spr_gfx_ram + (tidx * bpp * 8) / 2 + ((ty * 8 + tx) / (8 / bpp)) / 2;

      switch (bpp) {
        case 4:
          *tptr |= colour << ((x & 3) << 2);

          break;

        case 8:
          *tptr |= colour << ((x & 1) << 3);

          break;

        default:
          break;
      }
    }
  }
}

void nds_init_sprite(int bpp)
{
  int dim = (TILE_WIDTH > TILE_HEIGHT) ? TILE_WIDTH : TILE_HEIGHT;
  u16 *spr_palette = SPRITE_PALETTE;
  int i;
  int spr_size;

  /* First thing's f'ing last, let's disable all sprites */

  for (i = 0; i < 128; i++) {
    switch (dim) {
      case 8:
        oam_shadow.spriteBuffer[i].objSize = OBJSIZE_8;
        spr_size = 1;
        break;

      case 16:
        oam_shadow.spriteBuffer[i].objSize = OBJSIZE_16;
        spr_size = 2;
        break;

      case 24:
      case 32:
        oam_shadow.spriteBuffer[i].objSize = OBJSIZE_32;
        spr_size = 4;
        break;

      case 48:
      case 64:
        oam_shadow.spriteBuffer[i].objSize = OBJSIZE_64;
        spr_size = 8;
        break;

      default:
        break;
    }

    switch (bpp) {
      case 4:
        oam_shadow.spriteBuffer[i].colMode = OBJCOLOR_16;
        break;

      case 8:
        oam_shadow.spriteBuffer[i].colMode = OBJCOLOR_256;
        break;

      default:
        break;
    }

    oam_shadow.spriteBuffer[i].isHidden = 1;
    oam_shadow.spriteBuffer[i].isRotoscale = 0;
    oam_shadow.spriteBuffer[i].rsDouble = 0;
    oam_shadow.spriteBuffer[i].objMode = OBJMODE_BLENDED;
    oam_shadow.spriteBuffer[i].isMosaic = 0;
    oam_shadow.spriteBuffer[i].objShape = OBJSHAPE_SQUARE;

    oam_shadow.spriteBuffer[i].posX = 0;
    oam_shadow.spriteBuffer[i].posY = 0;
    oam_shadow.spriteBuffer[i].objPriority = OBJPRIORITY_3;
    oam_shadow.spriteBuffer[i].objPal = 0;

    if (i == 0) {
      oam_shadow.spriteBuffer[i].tileIdx = spr_size * spr_size * (bpp / 4);
    } else {
      oam_shadow.spriteBuffer[i].tileIdx = (spr_size * spr_size * (bpp / 4)) * 2;
    }
  }
  
  /* Let's draw our highlight thinger */

  spr_palette[1] = RGB15(31, 31, 31);
  spr_palette[2] = RGB15(0, 31, 0);

  nds_render_sprite(bpp, spr_size, 1, 1);
  nds_render_sprite(bpp, spr_size, 2, 2);
}

/*
 * Here we do two things.  First, we load the BMP into memory.  We throw an
 * error if it isn't an indexed file of some kind (bpp <= 8).  After the file 
 * is loaded, we populate the palette provided.  This involves converting the 
 * 24-bit RGB tuplets to 15-bit NDS palette entries.
 */
int nds_init_map()
{
  u16 *palette;
  int i;
  u16 blend_dst;

  int bpp;
  u16 palette_data[512];
  int palette_length;

  /* 
   * Alright, load the tile file or font data, depending on the rendering
   * mode.
   */

  if ((bpp = NDS_INIT_MAP(palette_data, &palette_length)) < 0) {
    return -1;
  }

  /* Initialize the data we need to manage the map and tiles */

  map_width = 32 / tile_width;
  map_height = 24 / tile_height;

  nds_init_tile_cache();

  /* Now initialize our graphics layer */

  /*
   * Set up the tile RAM starting at 16k.  At 65k, the bitmap data starts
   * for the menu layer, so we stick the map RAM 2k before that, 
   * which maximizes the amount of tile RAM we have.
   */

  switch (bpp) {
    case 4:
      BG1_CR = BG_32x32 | MAP_BASE | TILE_BASE | BG_16_COLOR | BG_PRIORITY_3; 
      DISPLAY_CR |= DISPLAY_BG1_ACTIVE;

      blend_dst = BLEND_DST_BG1;

      palette = (u16 *)BG_PALETTE + MAP_PALETTE_BASE;
      break;

    case 8:
      BG3_CR = BG_RS_32x32 | MAP_BASE | TILE_BASE | BG_PRIORITY_3; 
      DISPLAY_CR |= DISPLAY_BG3_ACTIVE;

      BG3_XDX = 1 << 8;
      BG3_XDY = 0;
      BG3_YDX = 0;
      BG3_YDY = 1 << 8;

      blend_dst = BLEND_DST_BG3;

      vramSetBankE(VRAM_E_LCD);
      palette = VRAM_E_EXT_PALETTE[3][2];

      break;

    default:
      iprintf("Sorry, %d bpp tile files aren't supported.\n", bmp_bpp(&tiles));

      return -1;
  }

  BLEND_CR = BLEND_ALPHA | BLEND_SRC_SPRITE | blend_dst;
  BLEND_AB = 0x0010;

  /* Alright, time to copy over the palette data. */

  for (i = 0; i < palette_length; i++) {
    palette[i] = palette_data[i];
  }

  /* If we're using extended palettes, get the VRAM set up. */

  if (bpp == 8) {
    vramSetBankE(VRAM_E_BG_EXT_PALETTE);
  }

  if (TILE_FILE != NULL) {
    nds_init_sprite(bpp);
  }

  return 0;
}

void nds_clear_map()
{
  memset(map_ram, 0, 32 * 24 * 2);
}

void nds_minimap_dims(int *x, int *y, int *x2, int *y2)
{
  int minimap_x = MINIMAP_X;
  int minimap_y = system_font->height + 2;

  /* Clear the edges of the map */

  *x = minimap_x - 1;
  *y = minimap_y - 1;
  *x2 = minimap_x + COLNO * 2;
  *y2 = minimap_y + ROWNO * 2;
}

void nds_draw_minimap(nds_map_t *map)
{
  u16 *sub_vram = (u16 *)BG_BMP_RAM_SUB(4);
  int x, y;
  int rx1, ry1, rx2, ry2;
  int bx1, by1, bx2, by2;

  nds_minimap_dims(&rx1, &ry1, &rx2, &ry2);

  nds_draw_hline(rx1, ry1, COLNO * 2 + 1, 0, sub_vram);
  nds_draw_hline(rx1, ry2, COLNO * 2 + 1, 0, sub_vram);

  nds_draw_vline(rx1, ry1, ROWNO * 2 + 1, 0, sub_vram);
  nds_draw_vline(rx2, ry1, ROWNO * 2 + 1, 0, sub_vram);

  for (y = 0; y < ROWNO; y++) {
    for (x = 0; x < COLNO; x++) {
      int glyph = map->glyphs[y][x];
      int typ = levl[x][y].typ;
      u16 colour;

      if (glyph < 0) {
        colour = 0x00;
      } else if ((x == u.ux) && (y == u.uy)) {
        colour = C_YOU;
      } else if (glyph_is_normal_monster(glyph)) {
        colour = C_MON; 
      } else if (glyph_is_pet(glyph)) {
        colour = C_PET;
      } else if (IS_WALL(typ)) {
        colour = C_WALL;
      } else if (IS_STWALL(typ)) {
        colour = 0x00;
      } else if (IS_DOOR(typ)) {
        colour = C_DOOR;
      } else if (typ == STAIRS) {
        colour = C_STAIRS;
      } else if (typ == ALTAR) {
        colour = C_ALTAR;
      } else if (typ == CORR) {
        colour = C_CORR;
      } else if (IS_FURNITURE(typ)) {
        colour = C_FURNITURE;
      } else if (IS_ROOM(typ)) {
        colour = C_ROOM;
      } else {
        colour = C_WALL;
      }

      sub_vram[(y + (ry1 + 1) / 2) * 256 + x + (rx1 + 1) / 2] = (colour << 8) | colour;
      sub_vram[(y + (ry1 + 1) / 2) * 256 + x + (rx1 + 1) / 2 + 128] = (colour << 8) | colour;
    }
  }

  bx1 = cx * 2 - map_width + rx1;
  bx2 = cx * 2 + map_width + rx1 + 1;

  by1 = cy * 2 - map_height + ry1;
  by2 = cy * 2 + map_height + ry1 + 1;

  nds_draw_hline(bx1, by1, map_width * 2 + 1, C_VISBORDER, sub_vram);
  nds_draw_hline(bx1, by2, map_width * 2 + 1, C_VISBORDER, sub_vram);
  nds_draw_vline(bx1, by1, map_height * 2 + 1, C_VISBORDER, sub_vram);
  nds_draw_vline(bx2, by1, map_height * 2 + 1, C_VISBORDER, sub_vram);
}

void nds_clear_minimap()
{
  u16 *sub_vram = (u16 *)BG_BMP_RAM_SUB(4);
  int x, y;
  int rx1, rx2, ry1, ry2;

  nds_minimap_dims(&rx1, &ry1, &rx2, &ry2);

  for (y = 0; y < ROWNO; y++) {
    for (x = 0; x < COLNO; x++) {
      sub_vram[(y + ry1 + 1) * 256 + x + rx1 + 1] = 0x0000;
      sub_vram[(y + ry1 + 1) * 256 + x + rx1 + 1 + 128] = 0x0000;
    }
  }

  nds_draw_hline(rx1 - 1, ry1 - 1, COLNO * 2 + 3, C_MAPBORDER, sub_vram);
  nds_draw_hline(rx1 - 1, ry2 + 1, COLNO * 2 + 3, C_MAPBORDER, sub_vram);

  nds_draw_vline(rx1 - 1, ry1 - 1, ROWNO * 2 + 3, C_MAPBORDER, sub_vram);
  nds_draw_vline(rx2 + 2, ry1 - 1, ROWNO * 2 + 3, C_MAPBORDER, sub_vram);
}

void nds_draw_map(nds_map_t *map, int *xp, int *yp)
{
  int sx, sy;

  swiWaitForVBlank();

  if (xp == NULL) {
    cx = u.ux;
  } else {
    cx = *xp;
  }
  
  if (yp == NULL) {
    cy = u.uy;
  } else {
    cy = *yp;
  }

  if ((cx + map_width / 2) > COLNO) {
    cx = COLNO - map_width / 2;
  } else if ((cx - map_width / 2) < 0) {
    cx = map_width / 2;
  }

  if ((cy + map_height / 2) > ROWNO) {
    cy = ROWNO - map_height / 2;
  } else if ((cy - map_height / 2) < 0) {
    cy = map_height / 2;
  }

  sx = cx - map_width / 2;
  sy = cy - map_height / 2;

  pet_count = 0;

  if (map != NULL) {
    int x, y;

    if (TILE_FILE != NULL) {
      nds_clear_sprites();
    } 

    for (y = 0; y < map_height; y++) {
      for (x = 0; x < map_width; x++) {
        if (((sx + x) < 0) || ((sx + x) >= COLNO) ||
            ((sy + y) < 0) || ((sy + y) >= ROWNO)) {
          nds_draw_tile(map, -1, x, y, sx + x, sy + y);
        } else {
          nds_draw_tile(map, map->glyphs[sy + y][sx + x], x, y, sx + x, sy + y);
        }
      }
    }

    if (TILE_FILE != NULL) {
      nds_draw_graphics_cursor(map->cx - sx, map->cy - sy);
    }

    nds_draw_minimap(map);
  } else {
    nds_draw_graphics_cursor(-1, -1);

    nds_clear_map();
    nds_clear_minimap();
  }

  DC_FlushAll();
  dmaCopy(&oam_shadow, oam_ram, sizeof(oam_shadow));
}

void nds_map_translate_coords(int x, int y, int *tx, int *ty)
{
  int sx = cx - map_width / 2;
  int sy = cy - map_height / 2;

  *tx = sx + x / (tile_width * 8);
  *ty = sy + y / (tile_height * 8);
}

int nds_map_tile_width()
{
  return tile_width * 8;
}

int nds_map_tile_height()
{
  return tile_height * 8;
}

void nds_map_get_center(int *xp, int *yp)
{
  *xp = cx;
  *yp = cy;
}

void nds_map_relativize(int *px, int *py)
{
  int upx, upy;
  int sx = cx - map_width / 2;
  int sy = cy - map_height / 2;
  int t_width = tile_width * 8;
  int t_height = tile_height * 8;

  upx = (u.ux - sx) * t_width + t_width / 2;
  upy = (u.uy - sy) * t_height + t_height / 2;

  *px -= upx;
  *py -= upy;
}
